package org.maachang.comet.net.nio ;

import java.net.ServerSocket;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.channels.Channel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Iterator;

import org.maachang.util.ArrayBinary;
import org.maachang.util.thread.LoopThread;

/**
 * NIO基本サーバ処理.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   MaachangComet 1.1B
 */
class NioBaseServer extends LoopThread {
    
    /**
     * 受信バッファサイズ.
     */
    private static final int BUFFER_LENGTH = 512 ;
    
    /**
     * サーバソケット.
     */
    private ServerSocket serverSocket = null ;
    
    /**
     * サーバソケットチャネル.
     */
    private ServerSocketChannel server = null ;
    
    /**
     * セレクタ.
     */
    private Selector selector = null ;
    
    /**
     * 受信処理キュー.
     */
    private ReceiveLinkQueue queue = null ;
    
    /**
     * 受信バッファ.
     */
    private ByteBuffer buffer = null ;
    
    /**
     * コンストラクタ.
     */
    private NioBaseServer() {
        
    }
    
    /**
     * コンストラクタ.
     * @param serverSocket 対象のサーバソケットを設定します.
     * @param array 受信実行スレッドとリンクするキューを設定します.
     * @exception Exception 例外.
     */
    public NioBaseServer( ServerSocket serverSocket,ReceiveLinkQueue queue )
        throws Exception {
        if( serverSocket == null || queue == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.queue = queue ;
        this.serverSocket = serverSocket ;
        this.serverSocket.setReuseAddress( true ) ;
        this.server = serverSocket.getChannel() ;
        this.selector = Selector.open() ;
        this.server.register( this.selector,SelectionKey.OP_ACCEPT ) ;
        this.buffer = ByteBuffer.allocateDirect( BUFFER_LENGTH ) ;
        this.startThread() ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        super.stopThread() ;
    }
    
    /**
     * オブジェクトクリア.
     */
    protected void clear() {
        if( this.selector != null ) {
            try {
                Iterator itr = this.selector.keys().iterator();
                while(itr.hasNext()){
                    try {
                        SelectionKey key = (SelectionKey)itr.next();
                        Channel c = key.channel();
                        c.close();
                    } catch( Exception ee ) {
                    }
                }
            } catch(Exception e) {}
        }
        this.selector = null ;
        if( this.server != null ) {
            try {
                this.server.close() ;
            } catch( Exception e ) {
            }
        }
        if( this.serverSocket != null ) {
            try {
                this.serverSocket.close() ;
            } catch( Exception e ) {
            }
        }
        this.serverSocket = null ;
        this.server = null ;
        this.queue = null ;
        this.buffer = null ;
    }
    
    /**
     * 実行処理.
     * @exception Exception 例外.
     */
    protected boolean execution() throws Exception {
        while ( selector.select() > 0 ) {
            Iterator itr = selector.selectedKeys().iterator() ;
            while(itr.hasNext()) {
                SelectionKey key = ( SelectionKey )itr.next() ;
                itr.remove() ;
                // accept.
                if( key.isAcceptable() ) {
                    SocketChannel sc = null ;
                    try {
                        ServerSocketChannel ssc = ( ServerSocketChannel )key.channel() ;
                        sc = ssc.accept() ;
                        if( sc != null ) {
                            setOption( sc ) ;
                            sc.register( selector,SelectionKey.OP_READ,new NioElement( sc ) ) ;
                        }
                    } catch( Exception e ) {
                        if( sc != null ) {
                            try { sc.close() ; } catch( Exception ee ) {}
                        }
                        continue ;
                    }
                }
                // use-socket.
                else if( key.isValid() ) {
                    boolean useRead = false ;
                    // read.
                    if( key.isReadable() ) {
                        NioElement em = null ;
                        try {
                            em = ( NioElement )key.attachment() ;
                            if( recv( em ) == false ) {
                                em.destroy() ;
                                continue ;
                            }
                            else {
                                if( key.interestOps() == SelectionKey.OP_READ ) {
                                    key.interestOps( SelectionKey.OP_READ | SelectionKey.OP_WRITE ) ;
                                }
                                useRead = true ;
                            }
                            if( em.isExecutionFlag() == false ) {
                                em.setExecutionFlag( true ) ;
                                queue.append( em ) ;
                            }
                        } catch( Exception e ) {
                            if( em != null ) {
                                em.destroy() ;
                            }
                            continue ;
                        }
                    }
                    // write.
                    if( key.isWritable() ) {
                        NioElement em = null ;
                        try {
                            em = ( NioElement )key.attachment() ;
                            ByteBuffer buf = em.getWriteBuffer().getQueue() ;
                            if( buf != null ) {
                                if( buf.limit() == 0 ) {
                                    if( em.isCloseFlag() == true ) {
                                        em.destroy() ;
                                    }
                                    else if( useRead == false ) {
                                        key.interestOps( SelectionKey.OP_READ ) ;
                                    }
                                }
                                else {
                                    SocketChannel ch = em.getChannel() ;
                                    ch.write( buf ) ;
                                }
                            }
                        } catch( Exception e ) {
                            if( em != null ) {
                                em.destroy() ;
                            }
                            continue ;
                        }
                    }
                }
                // not-use-socket.
                else {
                    try { (( NioElement )key.attachment()).destroy() ; } catch( Exception e ){}
                }
            }
        }
        return true ;
    }
    
    private static final int BUFFER = 131072 ;
    private static final int LINGER = 5 ;
    
    /**
     * Acceptソケット設定.
     */
    private void setOption( SocketChannel channel )
        throws Exception {
        channel.configureBlocking( false ) ;
        Socket soc = channel.socket() ;
        soc.setSendBufferSize( BUFFER ) ;
        soc.setReceiveBufferSize( BUFFER ) ;
        soc.setKeepAlive( true ) ;
        soc.setTcpNoDelay( true ) ;
        soc.setReuseAddress( true ) ;
        soc.setSoLinger( true,LINGER ) ;
    }
    
    /**
     * データ受信処理.
     */
    private boolean recv( NioElement emt ) throws Exception {
        ArrayBinary emtBuf = emt.getBuffer() ;
        SocketChannel channel = emt.getChannel() ;
        boolean flg = false ;
        for( ;; ) {
            buffer.clear() ;
            int len = channel.read( buffer ) ;
            if( len <= 0 ) {
                if( len <= -1 ) {
                    return false ;
                }
                break ;
            }
            buffer.flip() ;
            emtBuf.write( buffer ) ;
            flg = true ;
        }
        if( flg == true ) {
            emt.update() ;
        }
        return flg ;
    }
}
